@@ -10,9 +10,7 @@ class Grapher:
             This class is used to generate:
                     1) the graph (in dictionary form) { source_node: [destination_node1, destination_node2]}
                     2) the dataframe with relative_distances
-
     Inputs: The class consists of a pandas dataframe consisting of cordinates for bounding boxe and the image of the invoice/receipt.
-
     """
 
     def __init__(self, document):
@@ -20,6 +18,7 @@ class Grapher:
         self.nodes = document.nodes
 
 
+
     def graph_formation(self):
 
         """
@@ -31,7 +30,6 @@ class Grapher:
             Two words (W_a and W_b) are in same line if:
                 Top(W_a) <= Bottom(W_b) and Bottom(W_a) >= Top(W_b)
         3) Sort words in each line based on Left coordinate
-
         This ensures that words are read from top left corner of the image first,
         going line by line from left to right and at last the final bottom right word of the page is read.
 
@@ -41,7 +39,6 @@ class Grapher:
         returns:
             df with words arranged in orientation top to bottom and left to right, the line number for each word, index of the node connected to
             on all directions top, bottom, right and left (if they exist and satisfy the parameters provided)
-
         _____________________y axis______________________
         |
         |                       top
@@ -51,8 +48,6 @@ class Grapher:
         |                       bottom
         |
         |
-
-
         iterate through the rows twice to compare them.
         remember that the axes are inverted.
 
@@ -64,18 +59,17 @@ class Grapher:
         df = df[0].str.split(',', expand=True)
         temp = df.copy()
         temp[temp.columns] = temp.apply(lambda x: x.str.strip())
-        temp.columns = ['xmin', 'ymin', 'xmax', 'ymax','tag']
+        temp.columns = ['xmin', 'ymin', 'xmax', 'ymax', 'tag']
         temp[['xmin', 'ymin', 'xmax', 'ymax']] = temp[['xmin', 'ymin', 'xmax', 'ymax']].apply(pd.to_numeric)
         df = temp
 
 
 
-
         assert 'xmin' in df.columns, '"xmin" not in object map'
         assert 'xmax' in df.columns, '"xmax" not in object map'
         assert 'ymin' in df.columns, '"ymin" not in object map'
         assert 'ymax' in df.columns, '"ymax" not in object map'
-        assert 'tag' in df.columns, '"tag" column not in object map'
+        assert 'Object' in df.columns, '"Object" column not in object map'
 
         # remove empty spaces both in front and behind
         for col in df.columns:
@@ -93,7 +87,6 @@ class Grapher:
         # subtracting ymax by 1 to eliminate ambiguity of boxes being in both left and right
         df["ymax"] = df["ymax"].apply(lambda x: x - 1)
 
-
         master = []
         for idx, row in df.iterrows():
             # flatten the nested list
@@ -116,8 +109,6 @@ class Grapher:
                                 line.append(idx_2)
                 master.append(line)
 
-
-
         df2 = pd.DataFrame({'words_indices': master, 'line_number': [x for x in range(1, len(master) + 1)]})
         # explode the list columns eg : [1,2,3]
         df2 = df2.set_index('line_number').words_indices.apply(pd.Series).stack() \
@@ -223,22 +214,73 @@ class Grapher:
             if key in dic1: result.setdefault(key, []).append(dic1[key])
             if key in dic2: result.setdefault(key, []).append(dic2[key])
 
-
         G = nx.from_dict_of_lists(result)
 
-
         self.df = df
         return G, result, df
 
+        # features calculation
 
-    def relative_distance(self,image_width, image_height):
+    def get_text_features(self, df):
+        """
+        gets text features
+        Args: df
+        Returns: n_lower, n_upper, n_spaces, n_alpha, n_numeric,n_special
+        """
+        data = df['Object'].tolist()
+
+        '''
+            Args:
+                df
+
+            Returns: 
+                character and word features
+
+        '''
+        special_chars = ['&', '@', '#', '(', ')', '-', '+',
+                         '=', '*', '%', '.', ',', '\\', '/',
+                         '|', ':']
+
+        # character wise
+        n_lower, n_upper, n_spaces, n_alpha, n_numeric, n_special = [], [], [], [], [], []
+
+        for words in data:
+            lower, upper, alpha, spaces, numeric, special = 0, 0, 0, 0, 0, 0
+            for char in words:
+                if char.islower():
+                    lower += 1
+                # for upper letters
+                if char.isupper():
+                    upper += 1
+                # for white spaces
+                if char.isspace():
+                    spaces += 1
+                    # for alphabetic chars
+                if char.isalpha():
+                    alpha += 1
+                    # for numeric chars
+                if char.isnumeric():
+                    numeric += 1
+                if char in special_chars:
+                    special += 1
+
+            n_lower.append(lower)
+            n_upper.append(upper)
+            n_spaces.append(spaces)
+            n_alpha.append(alpha)
+            n_numeric.append(numeric)
+            n_special.append(special)
+
+
+        df['n_upper'], df['n_alpha'], df['n_spaces'], \
+        df['n_numeric'], df['n_special'] = n_upper, n_alpha, n_spaces, n_numeric, n_special
+
+    def relative_distance(self):
         """
         1) Calculates relative distances for each node in left, right, top  and bottom directions if they exist.
         rd_l, rd_r = relative distances left , relative distances right. The distances are divided by image width
         rd_t, rd_b = relative distances top , relative distances bottom. The distances are divided by image length
-
         2) Exports the complete document graph for visualization
-
         Args:
             result dataframe from graph_formation()
 
@@ -246,7 +288,8 @@ class Grapher:
             dataframe with features and exports document graph if prompted
         """
 
-        df = self.df
+        df, img = self.df, self.image
+        image_height, image_width = self.image.shape[0], self.image.shape[1]
 
 
         for index in df['index'].to_list():
@@ -298,12 +341,9 @@ class Grapher:
         # replace all tne NaN values with '0' meaning there is nothing in that direction
         df[['rd_r', 'rd_b', 'rd_l', 'rd_t']] = df[['rd_r', 'rd_b', 'rd_l', 'rd_t']].fillna(0)
 
-
         # drop the unnecessary columns
-        df.drop(['destination_x_hori', 'destination_y_hori', 'destination_y_vert', 'destination_x_vert', 'left', 'top', 'right', 'bottom'], axis=1,
+        df.drop(['destination_x_hori', 'destination_y_hori', 'destination_y_vert', 'destination_x_vert'], axis=1,
                 inplace=True)
-
-
+        self.get_text_features(df)
         return df
 
-
